@@ -4,7 +4,7 @@
  *
  * (C) 2018-2022 by Vadim Yanitskiy <axilirator@gmail.com>
  * (C) 2018 by Harald Welte <laforge@gnumonks.org>
- * Contributions by sysmocom - s.f.m.c. GmbH
+ * (C) 2020-2023 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
  *
  * All Rights Reserved
  *
@@ -31,6 +31,7 @@
 
 #include <osmocom/gsm/protocol/gsm_04_08.h>
 #include <osmocom/gsm/gsm_utils.h>
+#include <osmocom/gsm/gsm0502.h>
 
 #include <osmocom/coding/gsm0503_coding.h>
 #include <osmocom/coding/gsm0503_amr_dtx.h>
@@ -39,6 +40,16 @@
 #include <osmocom/bb/l1sched/l1sched.h>
 #include <osmocom/bb/l1sched/logging.h>
 
+/* Burst Payload LENgth (short alias) */
+#define BPLEN GSM_NBITS_NB_GMSK_PAYLOAD
+
+/* Burst BUFfer capacity (in BPLEN units) */
+#define BUFMAX 24
+
+/* Burst BUFfer position macros */
+#define BUFPOS(buf, n) &buf[(n) * BPLEN]
+#define BUFTAIL8(buf) BUFPOS(buf, (BUFMAX - 8))
+
 /* 3GPP TS 45.009, table 3.2.1.3-{2,4}: AMR on Downlink TCH/H.
  *
  * +---+---+---+---+---+---+
@@ -123,6 +134,41 @@ static const uint8_t sched_tchh_dl_facch_map[26] = {
 	[7]  = 1, /* FACCH/H(1): B2(22,24,1,3,5,7) */
 };
 
+/* 3GPP TS 45.002, table 2 in clause 7: Mapping tables for TCH/H2.4 and TCH/H4.8.
+ *
+ * +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
+ * | a | b | c | d | e | f | g | h | i | j | k | l | m | n | o | p | q | r | s | t | u | v |
+ * +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
+ *
+ * TCH/H(0): B0(0,2,4,6,8,10,13,15,17,19,21,23,0,2,4,6,8,10,13,15,17,19)
+ * TCH/H(1): B0(1,3,5,7,9,11,14,16,18,20,22,24,1,3,5,7,9,11,14,16,18,20)
+ * TCH/H(0): B1(8,10,13,15,17,19,21,23,0,2,4,6,8,10,13,15,17,19,21,23,0,2)
+ * TCH/H(1): B1(9,11,14,16,18,20,22,24,1,3,5,7,9,11,14,16,18,20,22,24,1,3)
+ * TCH/H(0): B2(17,19,21,23,0,2,4,6,8,10,13,15,17,19,21,23,0,2,4,6,8,10)
+ * TCH/H(1): B2(18,20,22,24,1,3,5,7,9,11,14,16,18,20,22,24,1,3,5,7,9,11)
+ *
+ * TDMA frame number of burst 'a' % 26 is the table index.
+ * This mapping is valid for both TCH/H(0) and TCH/H(1). */
+static const uint8_t sched_tchh_ul_csd_map[26] = {
+	[0]  = 1, /* TCH/H(0): B0(0  ... 19) */
+	[1]  = 1, /* TCH/H(1): B0(1  ... 20) */
+	[8]  = 1, /* TCH/H(0): B1(8  ... 2) */
+	[9]  = 1, /* TCH/H(1): B1(9  ... 3) */
+	[17] = 1, /* TCH/H(0): B2(17 ... 10) */
+	[18] = 1, /* TCH/H(1): B2(18 ... 11) */
+};
+
+/* TDMA frame number of burst 'v' % 26 is the table index.
+ * This mapping is valid for both TCH/H(0) and TCH/H(1). */
+static const uint8_t sched_tchh_dl_csd_map[26] = {
+	[19] = 1, /* TCH/H(0): B0(0  ... 19) */
+	[20] = 1, /* TCH/H(1): B0(1  ... 20) */
+	[2]  = 1, /* TCH/H(0): B1(8  ... 2) */
+	[3]  = 1, /* TCH/H(1): B1(9  ... 3) */
+	[10] = 1, /* TCH/H(0): B2(17 ... 10) */
+	[11] = 1, /* TCH/H(1): B2(18 ... 11) */
+};
+
 /**
  * Can a TCH/H block transmission be initiated / finished
  * on a given frame number and a given channel type?
@@ -180,78 +226,51 @@ bool l1sched_tchh_block_map_fn(enum l1sched_lchan_type chan,
 	return false;
 }
 
-/**
- * Calculates a frame number of the first burst
- * using given frame number of the last burst.
- *
- * See GSM 05.02, clause 7, table 1
- *
- * @param  chan      channel type (L1SCHED_TCHH_0 or L1SCHED_TCHH_1)
- * @param  last_fn   frame number of the last burst
- * @param  facch     FACCH/H or traffic?
- * @return           either frame number of the first burst,
- *                   or fn=last_fn if calculation failed
- */
-static uint32_t tchh_block_dl_first_fn(const struct l1sched_lchan_state *lchan,
-				       uint32_t last_fn, bool facch)
+static int decode_hr_facch(struct l1sched_lchan_state *lchan)
 {
-	enum l1sched_lchan_type chan = lchan->type;
-	uint8_t fn_mf, fn_diff;
-	int i = 0;
+	uint8_t data[GSM_MACBLOCK_LEN];
+	int n_errors, n_bits_total;
+	int rc;
 
-	/* Just to be sure */
-	OSMO_ASSERT(chan == L1SCHED_TCHH_0 || chan == L1SCHED_TCHH_1);
+	rc = gsm0503_tch_hr_facch_decode(&data[0], BUFTAIL8(lchan->rx_bursts),
+					 &n_errors, &n_bits_total);
+	if (rc != GSM_MACBLOCK_LEN)
+		return rc;
 
-	/* Calculate a modulo */
-	fn_mf = facch ? (last_fn % 26) : (last_fn % 13);
+	/* calculate AVG of the measurements (FACCH/H takes 6 bursts) */
+	l1sched_lchan_meas_avg(lchan, 6);
 
-#define BLOCK_FIRST_FN(map) \
-	do { \
-		if (map[i][ARRAY_SIZE(map[i]) - 1] == fn_mf) { \
-			fn_diff = GSM_TDMA_FN_DIFF(fn_mf, map[i][0]); \
-			return GSM_TDMA_FN_SUB(last_fn, fn_diff); \
-		} \
-	} while (++i < ARRAY_SIZE(map))
+	l1sched_lchan_emit_data_ind(lchan, &data[0], GSM_MACBLOCK_LEN,
+				    n_errors, n_bits_total, false);
 
-	/* Choose a proper block map */
-	if (facch) {
-		if (chan == L1SCHED_TCHH_0)
-			BLOCK_FIRST_FN(tch_h0_dl_facch_block_map);
-		else
-			BLOCK_FIRST_FN(tch_h1_dl_facch_block_map);
-	} else {
-		if (chan == L1SCHED_TCHH_0)
-			BLOCK_FIRST_FN(tch_h0_traffic_block_map);
-		else
-			BLOCK_FIRST_FN(tch_h1_traffic_block_map);
-	}
-
-	LOGP_LCHAND(lchan, LOGL_ERROR,
-		    "Failed to calculate TDMA frame number of the first burst of %s block, "
-		    "using the current fn=%u\n", facch ? "FACCH/H" : "TCH/H", last_fn);
-
-	/* Couldn't calculate the first fn, return the last */
-	return last_fn;
+	return GSM_MACBLOCK_LEN;
 }
 
 int rx_tchh_fn(struct l1sched_lchan_state *lchan,
 	       const struct l1sched_burst_ind *bi)
 {
 	int n_errors = -1, n_bits_total = 0, rc;
-	sbit_t *buffer, *offset;
-	uint8_t l2[128], *mask;
-	size_t l2_len;
+	sbit_t *bursts_p, *burst;
+	uint8_t tch_data[240];
+	size_t tch_data_len;
+	uint32_t *mask;
 	int amr = 0;
 	uint8_t ft;
-	bool fn_is_cmi;
 
 	/* Set up pointers */
 	mask = &lchan->rx_burst_mask;
-	buffer = lchan->rx_bursts;
+	bursts_p = lchan->rx_bursts;
 
 	LOGP_LCHAND(lchan, LOGL_DEBUG,
 		    "Traffic received: fn=%u bid=%u\n", bi->fn, bi->bid);
 
+	if (bi->bid == 0) {
+		/* Shift the burst buffer by 2 bursts leftwards */
+		memmove(BUFPOS(bursts_p, 0), BUFPOS(bursts_p, 2), 20 * BPLEN);
+		memset(BUFPOS(bursts_p, 20), 0, 2 * BPLEN);
+		*mask = *mask << 2;
+	}
+
 	if (*mask == 0x00) {
 		/* Align to the first burst */
 		if (bi->bid > 0)
@@ -261,9 +280,6 @@ int rx_tchh_fn(struct l1sched_lchan_state *lchan,
 		if (lchan->tch_mode == GSM48_CMODE_SIGN) {
 			if (!l1sched_tchh_facch_start(lchan->type, bi->fn, 0))
 				return 0;
-		} else { /* or TCH/H traffic frame */
-			if (!l1sched_tchh_traffic_start(lchan->type, bi->fn, 0))
-				return 0;
 		}
 	}
 
@@ -273,51 +289,69 @@ int rx_tchh_fn(struct l1sched_lchan_state *lchan,
 	/* Store the measurements */
 	l1sched_lchan_meas_push(lchan, bi);
 
-	/* Copy burst to the end of buffer of 6 bursts */
-	offset = buffer + bi->bid * 116 + 464;
-	memcpy(offset, bi->burst + 3, 58);
-	memcpy(offset + 58, bi->burst + 87, 58);
+	/* Copy burst to the end of buffer of 24 bursts */
+	burst = BUFPOS(bursts_p, 20 + bi->bid);
+	memcpy(burst, bi->burst + 3, 58);
+	memcpy(burst + 58, bi->burst + 87, 58);
 
 	/* Wait until the second burst */
 	if (bi->bid != 1)
 		return 0;
 
 	/* Wait for complete set of bursts */
-	if (lchan->tch_mode == GSM48_CMODE_SIGN) {
+	switch (lchan->tch_mode) {
+	case GSM48_CMODE_SIGN:
 		/* FACCH/H is interleaved over 6 bursts */
 		if ((*mask & 0x3f) != 0x3f)
-			goto bfi_shift;
-	} else {
-		/* Traffic is interleaved over 4 bursts */
+			return 0;
+		break;
+	case GSM48_CMODE_DATA_6k0:
+	case GSM48_CMODE_DATA_3k6:
+		/* Data (CSD) is interleaved over 22 bursts */
+		if ((*mask & 0x3fffff) != 0x3fffff)
+			return 0;
+		if (!sched_tchh_dl_csd_map[bi->fn % 26])
+			return 0; /* CSD: skip decoding attempt, need 2 more bursts */
+		break;
+	default:
+		/* Speech is interleaved over 4 bursts */
 		if ((*mask & 0x0f) != 0x0f)
-			goto bfi_shift;
+			return 0;
+		break;
 	}
 
 	/* Skip decoding attempt in case of FACCH/H */
 	if (lchan->dl_ongoing_facch) {
+		/* Send BFI (DATA.ind without payload) for the 2nd stolen TCH frame */
+		l1sched_lchan_meas_avg(lchan, 4);
+		l1sched_lchan_emit_data_ind(lchan, NULL, 0, 0, 0, true);
 		lchan->dl_ongoing_facch = false;
-		goto bfi_shift; /* 2/2 BFI */
+		return 0;
 	}
 
+	/* TCH/H: speech and signalling frames are interleaved over 4 and 6 bursts,
+	 * respectively, while CSD frames are interleaved over 22 bursts.  Unless
+	 * we're in CSD mode, decode only the last 6 bursts to avoid introducing
+	 * additional delays. */
 	switch (lchan->tch_mode) {
 	case GSM48_CMODE_SIGN:
 	case GSM48_CMODE_SPEECH_V1: /* HR */
-		rc = gsm0503_tch_hr_decode(l2, buffer,
-			!l1sched_tchh_facch_end(lchan->type, bi->fn, 0),
-			&n_errors, &n_bits_total);
+		rc = gsm0503_tch_hr_decode(&tch_data[0], BUFTAIL8(bursts_p),
+					   !sched_tchh_dl_facch_map[bi->fn % 26],
+					   &n_errors, &n_bits_total);
 		break;
 	case GSM48_CMODE_SPEECH_AMR: /* AMR */
-		/* the first FN FN 4,13,21 or 5,14,22 defines that CMI is
-		 * included in frame, the first FN FN 0,8,17 or 1,9,18 defines
-		 * that CMR/CMC is included in frame. */
-		fn_is_cmi = sched_tchh_dl_amr_cmi_map[bi->fn % 26];
-
 		/* See comment in function rx_tchf_fn() */
 		amr = 2;
-		rc = gsm0503_tch_ahs_decode_dtx(l2 + amr, buffer,
-			!sched_tchh_dl_facch_map[bi->fn % 26],
-			!fn_is_cmi, lchan->amr.codec, lchan->amr.codecs, &lchan->amr.dl_ft,
-			&lchan->amr.dl_cmr, &n_errors, &n_bits_total, &lchan->amr.last_dtx);
+		rc = gsm0503_tch_ahs_decode_dtx(&tch_data[amr], BUFTAIL8(bursts_p),
+						!sched_tchh_dl_facch_map[bi->fn % 26],
+						!sched_tchh_dl_amr_cmi_map[bi->fn % 26],
+						lchan->amr.codec,
+						lchan->amr.codecs,
+						&lchan->amr.dl_ft,
+						&lchan->amr.dl_cmr,
+						&n_errors, &n_bits_total,
+						&lchan->amr.last_dtx);
 
 		/* only good speech frames get rtp header */
 		if (rc != GSM_MACBLOCK_LEN && rc >= 4) {
@@ -327,7 +361,7 @@ int rx_tchh_fn(struct l1sched_lchan_state *lchan,
 				/* SID frames will always get Frame Type Index 8 (AMR_SID) */
 				ft = AMR_SID;
 			}
-			rc = osmo_amr_rtp_enc(l2,
+			rc = osmo_amr_rtp_enc(&tch_data[0],
 				lchan->amr.codec[lchan->amr.dl_cmr],
 				ft, AMR_GOOD);
 			if (rc < 0)
@@ -335,29 +369,33 @@ int rx_tchh_fn(struct l1sched_lchan_state *lchan,
 					    "osmo_amr_rtp_enc() returned rc=%d\n", rc);
 		}
 		break;
+	/* CSD (TCH/H4.8): 6.0 kbit/s radio interface rate */
+	case GSM48_CMODE_DATA_6k0:
+		/* FACCH/H does not steal TCH/H4.8 frames, but only disturbs some bits */
+		decode_hr_facch(lchan);
+		rc = gsm0503_tch_hr48_decode(&tch_data[0], BUFPOS(bursts_p, 0),
+					     &n_errors, &n_bits_total);
+		break;
+	/* CSD (TCH/H2.4): 3.6 kbit/s radio interface rate */
+	case GSM48_CMODE_DATA_3k6:
+		/* FACCH/H does not steal TCH/H2.4 frames, but only disturbs some bits */
+		decode_hr_facch(lchan);
+		rc = gsm0503_tch_hr24_decode(&tch_data[0], BUFPOS(bursts_p, 0),
+					     &n_errors, &n_bits_total);
+		break;
 	default:
 		LOGP_LCHAND(lchan, LOGL_ERROR, "Invalid TCH mode: %u\n", lchan->tch_mode);
 		return -EINVAL;
 	}
 
-	/* Shift buffer by 4 bursts for interleaving */
-	memcpy(buffer, buffer + 232, 232);
-	memcpy(buffer + 232, buffer + 464, 232);
-
-	/* Shift burst mask */
-	*mask = *mask << 2;
-
 	/* Check decoding result */
 	if (rc < 4) {
-		/* Calculate AVG of the measurements (assuming 4 bursts) */
-		l1sched_lchan_meas_avg(lchan, 4);
-
 		LOGP_LCHAND(lchan, LOGL_ERROR,
 			    "Received bad frame (rc=%d, ber=%d/%d) at fn=%u\n",
 			    rc, n_errors, n_bits_total, lchan->meas_avg.fn);
 
-		/* Send BFI */
-		goto bfi;
+		/* Send BFI (DATA.ind without payload) */
+		tch_data_len = 0;
 	} else if (rc == GSM_MACBLOCK_LEN) {
 		/* Skip decoding of the next 2 stolen bursts */
 		lchan->dl_ongoing_facch = true;
@@ -366,197 +404,191 @@ int rx_tchh_fn(struct l1sched_lchan_state *lchan,
 		l1sched_lchan_meas_avg(lchan, 6);
 
 		/* FACCH/H received, forward to the higher layers */
-		l1sched_handle_data_ind(lchan, l2 + amr, GSM_MACBLOCK_LEN,
-					n_errors, n_bits_total,
-					L1SCHED_DT_SIGNALING);
+		l1sched_lchan_emit_data_ind(lchan, &tch_data[amr], GSM_MACBLOCK_LEN,
+					    n_errors, n_bits_total, false);
 
-		/* Send BFI substituting 1/2 stolen TCH frames */
-		n_errors = -1; /* ensure fake measurements */
-		goto bfi;
+		/* Send BFI (DATA.ind without payload) for the 1st stolen TCH frame */
+		if (lchan->tch_mode == GSM48_CMODE_SIGN)
+			return 0;
+		tch_data_len = 0;
 	} else {
 		/* A good TCH frame received */
-		l2_len = rc;
-
-		/* Calculate AVG of the measurements (traffic takes 4 bursts) */
-		l1sched_lchan_meas_avg(lchan, 4);
+		tch_data_len = rc;
 	}
 
+	/* Calculate AVG of the measurements (traffic takes 4 bursts) */
+	l1sched_lchan_meas_avg(lchan, 4);
+
 	/* Send a traffic frame to the higher layers */
-	return l1sched_handle_data_ind(lchan, l2, l2_len,
-				       n_errors, n_bits_total,
-				       L1SCHED_DT_TRAFFIC);
-
-bfi_shift:
-	/* Shift buffer */
-	memcpy(buffer, buffer + 232, 232);
-	memcpy(buffer + 232, buffer + 464, 232);
-
-	/* Shift burst mask */
-	*mask = *mask << 2;
-
-bfi:
-	/* Didn't try to decode, fake measurements */
-	if (n_errors < 0) {
-		lchan->meas_avg = (struct l1sched_meas_set) {
-			.fn = tchh_block_dl_first_fn(lchan, bi->fn, false),
-			.toa256 = 0,
-			.rssi = -110,
-		};
-
-		/* No bursts => no errors */
-		n_errors = 0;
-	}
-
-	/* BFI is not applicable in signalling mode */
-	if (lchan->tch_mode == GSM48_CMODE_SIGN) {
-		return l1sched_handle_data_ind(lchan, NULL, 0,
-					       n_errors, n_bits_total,
-					       L1SCHED_DT_SIGNALING);
-	}
-
-	/* Bad frame indication */
-	l2_len = l1sched_bad_frame_ind(l2, lchan);
-
-	/* Send a BFI frame to the higher layers */
-	return l1sched_handle_data_ind(lchan, l2, l2_len,
-				       n_errors, n_bits_total,
-				       L1SCHED_DT_TRAFFIC);
+	return l1sched_lchan_emit_data_ind(lchan, &tch_data[0], tch_data_len,
+					   n_errors, n_bits_total, true);
 }
 
 int tx_tchh_fn(struct l1sched_lchan_state *lchan,
 	       struct l1sched_burst_req *br)
 {
-	ubit_t *buffer, *offset;
+	struct msgb *msg_facch, *msg_tch, *msg;
+	ubit_t *bursts_p, *burst;
 	const uint8_t *tsc;
-	uint8_t *mask;
-	size_t l2_len;
+	uint32_t *mask;
 	int rc;
 
 	/* Set up pointers */
 	mask = &lchan->tx_burst_mask;
-	buffer = lchan->tx_bursts;
+	bursts_p = lchan->tx_bursts;
 
 	if (br->bid > 0) {
-		/* Align to the first burst */
-		if (*mask == 0x00)
-			return 0;
+		if ((*mask & 0x01) != 0x01)
+			return -ENOENT;
 		goto send_burst;
 	}
 
 	if (*mask == 0x00) {
-		/* Align transmission of the first FACCH/H frame */
-		if (lchan->tch_mode == GSM48_CMODE_SIGN)
+		/* Align transmission of the first frame */
+		switch (lchan->tch_mode) {
+		case GSM48_CMODE_SIGN:
 			if (!l1sched_tchh_facch_start(lchan->type, br->fn, 1))
 				return 0;
+			break;
+		case GSM48_CMODE_DATA_6k0:
+		case GSM48_CMODE_DATA_3k6:
+			if (!sched_tchh_ul_csd_map[br->fn % 26])
+				return 0;
+			break;
+		}
 	}
 
-	/* Shift buffer by 2 bursts back for interleaving */
-	memcpy(buffer, buffer + 232, 232);
-
-	/* Also shift TX burst mask */
+	/* Shift the burst buffer by 2 bursts leftwards for interleaving */
+	memmove(BUFPOS(bursts_p, 0), BUFPOS(bursts_p, 2), 20 * BPLEN);
+	memset(BUFPOS(bursts_p, 20), 0, 2 * BPLEN);
 	*mask = *mask << 2;
 
 	/* If FACCH/H blocks are still pending */
 	if (lchan->ul_facch_blocks > 2) {
-		memcpy(buffer + 232, buffer + 464, 232);
+		struct msgb *msg = l1sched_lchan_prim_dequeue_tch(lchan, false);
+		msgb_free(msg); /* drop 2nd TCH/HS block */
 		goto send_burst;
 	}
 
-	/* populate the buffer with bursts */
-	if (L1SCHED_PRIM_IS_FACCH(lchan->prim)) {
-		rc = gsm0503_tch_hr_encode(buffer, lchan->prim->payload,  lchan->prim->payload_len);
-		lchan->ul_facch_blocks = 6;
-	} else if (lchan->tch_mode == GSM48_CMODE_SPEECH_AMR) {
-		int len;
-		uint8_t cmr_codec;
-		int ft, cmr, i;
-		enum osmo_amr_type ft_codec;
-		enum osmo_amr_quality bfi;
-		int8_t sti, cmi;
-		bool amr_fn_is_cmr;
-		/* the first FN 0,8,17 defines that CMI is included in frame,
-		 * the first FN 4,13,21 defines that CMR is included in frame.
-		 */
-		amr_fn_is_cmr = !sched_tchh_ul_amr_cmi_map[br->fn % 26];
+	switch (lchan->tch_mode) {
+	case GSM48_CMODE_DATA_6k0:
+	case GSM48_CMODE_DATA_3k6:
+		/* CSD: skip dequeueing/encoding, send 2 more bursts */
+		if (!sched_tchh_ul_csd_map[br->fn % 26])
+			goto send_burst;
+		break;
+	}
 
-		len = osmo_amr_rtp_dec(lchan->prim->payload, lchan->prim->payload_len,
-				&cmr_codec, &cmi, &ft_codec,
-				&bfi, &sti);
-		if (len < 0) {
-			LOGP_LCHAND(lchan, LOGL_ERROR, "Cannot send invalid AMR payload (%zu): %s\n",
-				lchan->prim->payload_len, osmo_hexdump(lchan->prim->payload, lchan->prim->payload_len));
-			goto free_bad_msg;
+	/* dequeue a pair of TCH and FACCH frames */
+	msg_tch = l1sched_lchan_prim_dequeue_tch(lchan, false);
+	if (l1sched_tchh_facch_start(lchan->type, br->fn, true))
+		msg_facch = l1sched_lchan_prim_dequeue_tch(lchan, true);
+	else
+		msg_facch = NULL;
+	/* prioritize FACCH over TCH */
+	msg = (msg_facch != NULL) ? msg_facch : msg_tch;
+
+	/* populate the buffer with bursts */
+	switch (lchan->tch_mode) {
+	case GSM48_CMODE_SIGN:
+		if (!l1sched_tchh_facch_start(lchan->type, br->fn, 1))
+			goto send_burst; /* XXX: should not happen */
+		if (msg == NULL)
+			msg = l1sched_lchan_prim_dummy_lapdm(lchan);
+		/* fall-through */
+	case GSM48_CMODE_SPEECH_V1:
+		if (msg == NULL) {
+			/* transmit a dummy speech block with inverted CRC3 */
+			gsm0503_tch_hr_encode(bursts_p, NULL, 0);
+			goto send_burst;
 		}
-		ft = -1;
-		cmr = -1;
-		for (i = 0; i < lchan->amr.codecs; i++) {
-			if (lchan->amr.codec[i] == ft_codec)
-				ft = i;
-			if (lchan->amr.codec[i] == cmr_codec)
-				cmr = i;
+		rc = gsm0503_tch_hr_encode(BUFPOS(bursts_p, 0),
+					   msgb_l2(msg),
+					   msgb_l2len(msg));
+		break;
+	case GSM48_CMODE_SPEECH_AMR:
+	{
+		bool amr_fn_is_cmr = !sched_tchh_ul_amr_cmi_map[br->fn % 26];
+		const uint8_t *data = msg ? msgb_l2(msg) : NULL;
+		size_t data_len = msg ? msgb_l2len(msg) : 0;
+
+		if (msg == NULL) {
+			/* TODO: It's not clear what to do for TCH/AHS.
+			 * TODO: Send dummy FACCH maybe? */
+			goto send_burst; /* send garbage */
 		}
-		if (ft < 0) {
-			LOGP_LCHAND(lchan, LOGL_ERROR,
-				    "Codec (FT = %d) of RTP frame not in list\n", ft_codec);
-			goto free_bad_msg;
+
+		if (data_len != GSM_MACBLOCK_LEN) { /* TCH/AHS: speech */
+			if (!l1sched_lchan_amr_prim_is_valid(lchan, msg, amr_fn_is_cmr))
+				goto free_bad_msg;
+			/* pull the AMR header - sizeof(struct amr_hdr) */
+			data_len -= 2;
+			data += 2;
 		}
-		if (amr_fn_is_cmr && lchan->amr.ul_ft != ft) {
-			LOGP_LCHAND(lchan, LOGL_ERROR,
-				    "Codec (FT = %d) of RTP cannot be changed now, but in next frame\n",
-				    ft_codec);
-			goto free_bad_msg;
+
+		rc = gsm0503_tch_ahs_encode(BUFPOS(bursts_p, 0),
+					    data, data_len,
+					    amr_fn_is_cmr,
+					    lchan->amr.codec,
+					    lchan->amr.codecs,
+					    lchan->amr.ul_ft,
+					    lchan->amr.ul_cmr);
+		break;
+	}
+	/* CSD (TCH/H4.8): 6.0 kbit/s radio interface rate */
+	case GSM48_CMODE_DATA_6k0:
+		if ((msg = msg_tch) != NULL) {
+			OSMO_ASSERT(msgb_l2len(msg) == 4 * 60);
+			gsm0503_tch_hr48_encode(BUFPOS(bursts_p, 0), msgb_l2(msg));
+			/* Confirm data sending (pass ownership of the msgb/prim) */
+			l1sched_lchan_emit_data_cnf(lchan, msg, br->fn);
+		} /* else: all bits of this frame are set to zero */
+		if ((msg = msg_facch) != NULL) {
+			gsm0503_tch_hr_facch_encode(BUFPOS(bursts_p, 0), msgb_l2(msg));
+			/* Confirm FACCH sending (pass ownership of the msgb/prim) */
+			l1sched_lchan_emit_data_cnf(lchan, msg, br->fn);
 		}
-		lchan->amr.ul_ft = ft;
-		if (cmr < 0) {
-			LOGP_LCHAND(lchan, LOGL_ERROR,
-				    "Codec (CMR = %d) of RTP frame not in list\n", cmr_codec);
-		} else {
-			lchan->amr.ul_cmr = cmr;
+		goto send_burst;
+	/* CSD (TCH/H2.4): 3.6 kbit/s radio interface rate */
+	case GSM48_CMODE_DATA_3k6:
+		if ((msg = msg_tch) != NULL) {
+			OSMO_ASSERT(msgb_l2len(msg) == 4 * 36);
+			gsm0503_tch_hr24_encode(BUFPOS(bursts_p, 0), msgb_l2(msg));
+			/* Confirm data sending (pass ownership of the msgb/prim) */
+			l1sched_lchan_emit_data_cnf(lchan, msg, br->fn);
+		} /* else: all bits of this frame are set to zero */
+		if ((msg = msg_facch) != NULL) {
+			gsm0503_tch_hr_facch_encode(BUFPOS(bursts_p, 0), msgb_l2(msg));
+			/* Confirm FACCH sending (pass ownership of the msgb/prim) */
+			l1sched_lchan_emit_data_cnf(lchan, msg, br->fn);
 		}
-		rc = gsm0503_tch_ahs_encode(buffer, lchan->prim->payload + 2,
-			lchan->prim->payload_len - 2, amr_fn_is_cmr,
-			lchan->amr.codec, lchan->amr.codecs,
-			lchan->amr.ul_ft,
-			lchan->amr.ul_cmr);
-	} else {
-		/* Determine and check the payload length */
-		switch (lchan->tch_mode) {
-		case GSM48_CMODE_SIGN:
-		case GSM48_CMODE_SPEECH_V1: /* HR */
-			l2_len = GSM_HR_BYTES + 1;
-			break;
-		default:
-			LOGP_LCHAND(lchan, LOGL_ERROR,
-				    "Invalid TCH mode: %u, dropping frame...\n",
-				    lchan->tch_mode);
-			/* Forget this primitive */
-			l1sched_prim_drop(lchan);
-			return -EINVAL;
-		}
-		if (lchan->prim->payload_len != l2_len) {
-			LOGP_LCHAND(lchan, LOGL_ERROR, "Primitive has odd length %zu "
-				    "(expected %zu for TCH or %u for FACCH), so dropping...\n",
-				    lchan->prim->payload_len, l2_len, GSM_MACBLOCK_LEN);
-			/* Forget this primitive */
-			l1sched_prim_drop(lchan);
-			return -EINVAL;
-		}
-		rc = gsm0503_tch_hr_encode(buffer, lchan->prim->payload, l2_len);
+		goto send_burst;
+	default:
+		LOGP_LCHAND(lchan, LOGL_ERROR,
+			    "TCH mode %s is unknown or not supported\n",
+			    gsm48_chan_mode_name(lchan->tch_mode));
+		goto free_bad_msg;
 	}
 
 	if (rc) {
-		LOGP_LCHAND(lchan, LOGL_ERROR, "Failed to encode L2 payload (len=%zu): %s\n",
-			    lchan->prim->payload_len, osmo_hexdump(lchan->prim->payload,
-								   lchan->prim->payload_len));
+		LOGP_LCHAND(lchan, LOGL_ERROR, "Failed to encode L2 payload (len=%u): %s\n",
+			    msgb_l2len(msg), msgb_hexdump_l2(msg));
 free_bad_msg:
-		/* Forget this primitive */
-		l1sched_prim_drop(lchan);
+		msgb_free(msg_facch);
+		msgb_free(msg_tch);
 		return -EINVAL;
 	}
 
+	if (msgb_l2len(msg) == GSM_MACBLOCK_LEN)
+		lchan->ul_facch_blocks = 6;
+
+	/* Confirm data / traffic sending (pass ownership of the msgb/prim) */
+	l1sched_lchan_emit_data_cnf(lchan, msg, br->fn);
+	msgb_free((msg == msg_facch) ? msg_tch : msg_facch);
+
 send_burst:
 	/* Determine which burst should be sent */
-	offset = buffer + br->bid * 116;
+	burst = BUFPOS(bursts_p, br->bid);
 
 	/* Update mask */
 	*mask |= (1 << br->bid);
@@ -566,11 +598,11 @@ send_burst:
 
 	/* Compose a new burst */
 	memset(br->burst, 0, 3); /* TB */
-	memcpy(br->burst + 3, offset, 58); /* Payload 1/2 */
+	memcpy(br->burst + 3, burst, 58); /* Payload 1/2 */
 	memcpy(br->burst + 61, tsc, 26); /* TSC */
-	memcpy(br->burst + 87, offset + 58, 58); /* Payload 2/2 */
+	memcpy(br->burst + 87, burst + 58, 58); /* Payload 2/2 */
 	memset(br->burst + 145, 0, 3); /* TB */
-	br->burst_len = GSM_BURST_LEN;
+	br->burst_len = GSM_NBITS_NB_GMSK_BURST;
 
 	LOGP_LCHAND(lchan, LOGL_DEBUG, "Scheduled fn=%u burst=%u\n", br->fn, br->bid);
 
@@ -578,20 +610,5 @@ send_burst:
 	if (lchan->ul_facch_blocks)
 		lchan->ul_facch_blocks--;
 
-	if ((*mask & 0x0f) == 0x0f) {
-		/**
-		 * If no more FACCH/H blocks pending,
-		 * confirm data / traffic sending
-		 */
-		if (!lchan->ul_facch_blocks) {
-			enum l1sched_data_type dt = L1SCHED_PRIM_IS_TCH(lchan->prim) ?
-							L1SCHED_DT_TRAFFIC : L1SCHED_DT_SIGNALING;
-			l1sched_handle_data_cnf(lchan, br->fn, dt);
-		}
-
-		/* Forget processed primitive */
-		l1sched_prim_drop(lchan);
-	}
-
 	return 0;
 }
