@@ -68,18 +68,20 @@ static const uint8_t meas_rep_dummy[] = {
 	0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 };
 
-static int l1sched_cfg_pchan_comb_req(struct l1sched_state *sched,
+static int l1sched_cfg_pchan_comb_ind(struct l1sched_state *sched,
 				      uint8_t tn, enum gsm_phys_chan_config pchan)
 {
-	const struct l1sched_config_req cr = {
-		.type = L1SCHED_CFG_PCHAN_COMB,
-		.pchan_comb = {
-			.tn = tn,
-			.pchan = pchan,
-		},
-	};
+	struct l1sched_prim *prim;
+	struct msgb *msg;
 
-	return l1sched_handle_config_req(sched, &cr);
+	msg = l1sched_prim_alloc(L1SCHED_PRIM_T_PCHAN_COMB, PRIM_OP_INDICATION);
+	OSMO_ASSERT(msg != NULL);
+
+	prim = l1sched_prim_from_msgb(msg);
+	prim->pchan_comb_ind.tn = tn;
+	prim->pchan_comb_ind.pchan = pchan;
+
+	return l1sched_prim_to_user(sched, msg);
 }
 
 static void l1sched_a5_burst_enc(struct l1sched_lchan_state *lchan,
@@ -115,26 +117,14 @@ void l1sched_pull_burst(struct l1sched_state *sched, struct l1sched_burst_req *b
 		return;
 
 	/* Make sure that lchan is allocated and active */
-	lchan = l1sched_find_lchan(ts, chan);
+	lchan = l1sched_find_lchan_by_type(ts, chan);
 	if (lchan == NULL || !lchan->active)
 		return;
 
-	/* If no primitive is being processed, try obtaining one from Tx queue */
-	if (lchan->prim == NULL)
-		lchan->prim = l1sched_prim_dequeue(&ts->tx_prims, br->fn, lchan);
-	if (lchan->prim == NULL) {
-		/* If CBTX (Continuous Burst Transmission) is required */
-		if (l1sched_lchan_desc[chan].flags & L1SCHED_CH_FLAG_CBTX)
-			l1sched_prim_dummy(lchan);
-		if (lchan->prim == NULL)
-			return;
-	}
-
-	/* TODO: report TX buffers health to the higher layers */
-
 	/* Handover RACH needs to be handled regardless of the
 	 * current channel type and the associated handler. */
-	if (L1SCHED_PRIM_IS_RACH(lchan->prim) && lchan->prim->chan != L1SCHED_RACH)
+	struct msgb *msg = llist_first_entry_or_null(&lchan->tx_prims, struct msgb, list);
+	if (msg && l1sched_prim_type_from_msgb(msg) == L1SCHED_PRIM_T_RACH)
 		handler = l1sched_lchan_desc[L1SCHED_RACH].tx_fn;
 
 	/* Poke lchan handler */
@@ -145,27 +135,6 @@ void l1sched_pull_burst(struct l1sched_state *sched, struct l1sched_burst_req *b
 		l1sched_a5_burst_enc(lchan, br);
 }
 
-/* Pull *and send* Uplink bursts for all timeslots and the current TDMA Fn. */
-void l1sched_pull_send_frame(struct l1sched_state *sched)
-{
-	/* Advance TDMA frame number in order to give the transceiver
-	 * more time to handle the burst before the actual transmission. */
-	const uint32_t fn = GSM_TDMA_FN_SUM(sched->fn_counter_proc,
-					    sched->fn_counter_advance);
-
-	/* Iterate over timeslot list */
-	for (unsigned int tn = 0; tn < ARRAY_SIZE(sched->ts); tn++) {
-		struct l1sched_burst_req br = {
-			.fn = fn,
-			.tn = tn,
-			.burst_len = 0, /* NOPE.ind */
-		};
-
-		l1sched_pull_burst(sched, &br);
-		l1sched_handle_burst_req(sched, &br);
-	}
-}
-
 void l1sched_logging_init(int log_cat_common, int log_cat_data)
 {
 	l1sched_log_cat_common = log_cat_common;
@@ -181,8 +150,6 @@ struct l1sched_state *l1sched_alloc(void *ctx, const struct l1sched_cfg *cfg, vo
 		return NULL;
 
 	*sched = (struct l1sched_state) {
-		/* .clock_timer is set up in l1sched_clck_correct() */
-		.fn_counter_advance = cfg->fn_advance,
 		.priv = priv,
 	};
 
@@ -209,7 +176,6 @@ void l1sched_free(struct l1sched_state *sched)
 	for (tn = 0; tn < ARRAY_SIZE(sched->ts); tn++)
 		l1sched_del_ts(sched, tn);
 
-	l1sched_clck_reset(sched);
 	talloc_free(sched);
 }
 
@@ -227,10 +193,6 @@ void l1sched_reset(struct l1sched_state *sched, bool reset_clock)
 	for (tn = 0; tn < ARRAY_SIZE(sched->ts); tn++)
 		l1sched_del_ts(sched, tn);
 
-	/* Stop and reset clock counter if required */
-	if (reset_clock)
-		l1sched_clck_reset(sched);
-
 	memcpy(&sched->sacch_cache[0], &meas_rep_dummy[0], sizeof(meas_rep_dummy));
 }
 
@@ -272,15 +234,12 @@ void l1sched_del_ts(struct l1sched_state *sched, int tn)
 		talloc_free(lchan);
 	}
 
-	/* Flush queue primitives for TX */
-	l1sched_prim_flush_queue(&ts->tx_prims);
-
 	/* Remove ts from list and free memory */
 	sched->ts[tn] = NULL;
 	talloc_free(ts);
 
 	/* Notify transceiver about that */
-	l1sched_cfg_pchan_comb_req(sched, tn, GSM_PCHAN_NONE);
+	l1sched_cfg_pchan_comb_ind(sched, tn, GSM_PCHAN_NONE);
 }
 
 #define LAYOUT_HAS_LCHAN(layout, lchan) \
@@ -316,8 +275,6 @@ int l1sched_configure_ts(struct l1sched_state *sched, int tn,
 		    "(Re)configure TDMA timeslot #%u as %s\n",
 		    tn, ts->mf_layout->name);
 
-	/* Init queue primitives for TX */
-	INIT_LLIST_HEAD(&ts->tx_prims);
 	/* Init logical channels list */
 	INIT_LLIST_HEAD(&ts->lchans);
 
@@ -337,6 +294,9 @@ int l1sched_configure_ts(struct l1sched_state *sched, int tn,
 		/* Set channel type */
 		lchan->type = type;
 
+		/* Init the Tx queue */
+		INIT_LLIST_HEAD(&lchan->tx_prims);
+
 		/* Add to the list of channel states */
 		llist_add_tail(&lchan->list, &ts->lchans);
 
@@ -346,7 +306,7 @@ int l1sched_configure_ts(struct l1sched_state *sched, int tn,
 	}
 
 	/* Notify transceiver about TS activation */
-	l1sched_cfg_pchan_comb_req(sched, tn, config);
+	l1sched_cfg_pchan_comb_ind(sched, tn, config);
 
 	return 0;
 }
@@ -364,9 +324,6 @@ int l1sched_reset_ts(struct l1sched_state *sched, int tn)
 	/* Undefine multiframe layout */
 	ts->mf_layout = NULL;
 
-	/* Flush queue primitives for TX */
-	l1sched_prim_flush_queue(&ts->tx_prims);
-
 	/* Deactivate all logical channels */
 	l1sched_deactivate_all_lchans(ts);
 
@@ -377,7 +334,7 @@ int l1sched_reset_ts(struct l1sched_state *sched, int tn)
 	}
 
 	/* Notify transceiver about that */
-	l1sched_cfg_pchan_comb_req(sched, tn, GSM_PCHAN_NONE);
+	l1sched_cfg_pchan_comb_ind(sched, tn, GSM_PCHAN_NONE);
 
 	return 0;
 }
@@ -413,18 +370,40 @@ int l1sched_start_ciphering(struct l1sched_ts *ts, uint8_t algo,
 	return 0;
 }
 
-struct l1sched_lchan_state *l1sched_find_lchan(struct l1sched_ts *ts,
-	enum l1sched_lchan_type chan)
+struct l1sched_lchan_state *l1sched_find_lchan_by_type(struct l1sched_ts *ts,
+						       enum l1sched_lchan_type type)
 {
 	struct l1sched_lchan_state *lchan;
 
 	llist_for_each_entry(lchan, &ts->lchans, list)
-		if (lchan->type == chan)
+		if (lchan->type == type)
 			return lchan;
 
 	return NULL;
 }
 
+struct l1sched_lchan_state *l1sched_find_lchan_by_chan_nr(struct l1sched_state *sched,
+							  uint8_t chan_nr, uint8_t link_id)
+{
+	const struct l1sched_ts *ts = sched->ts[chan_nr & 0x07];
+	const struct l1sched_lchan_desc *lchan_desc;
+	struct l1sched_lchan_state *lchan;
+
+	if (ts == NULL)
+		return NULL;
+
+	llist_for_each_entry(lchan, &ts->lchans, list) {
+		lchan_desc = &l1sched_lchan_desc[lchan->type];
+		if (lchan_desc->chan_nr != (chan_nr & RSL_CHAN_NR_MASK))
+			continue;
+		if (lchan_desc->link_id != link_id)
+			continue;
+		return lchan;
+	}
+
+	return NULL;
+}
+
 int l1sched_set_lchans(struct l1sched_ts *ts, uint8_t chan_nr,
 		       int active, uint8_t tch_mode, uint8_t tsc)
 {
@@ -439,7 +418,7 @@ int l1sched_set_lchans(struct l1sched_ts *ts, uint8_t chan_nr,
 	llist_for_each_entry(lchan, &ts->lchans, list) {
 		lchan_desc = &l1sched_lchan_desc[lchan->type];
 
-		if (lchan_desc->chan_nr == (chan_nr & 0xf8)) {
+		if (lchan_desc->chan_nr == (chan_nr & RSL_CHAN_NR_MASK)) {
 			if (active) {
 				rc |= l1sched_activate_lchan(ts, lchan->type);
 				lchan->tch_mode = tch_mode;
@@ -452,13 +431,40 @@ int l1sched_set_lchans(struct l1sched_ts *ts, uint8_t chan_nr,
 	return rc;
 }
 
+int l1sched_lchan_set_amr_cfg(struct l1sched_lchan_state *lchan,
+			      uint8_t codecs_bitmask, uint8_t start_codec)
+{
+	int n = 0;
+	int acum = 0;
+	int pos;
+
+	while ((pos = ffs(codecs_bitmask)) != 0) {
+		acum += pos;
+		LOGP_LCHANC(lchan, LOGL_DEBUG, "AMR codec[%u] = %u\n", n, acum - 1);
+		lchan->amr.codec[n++] = acum - 1;
+		codecs_bitmask >>= pos;
+	}
+	if (n == 0) {
+		LOGP_LCHANC(lchan, LOGL_ERROR, "Empty AMR codec mode bitmask!\n");
+		return -EINVAL;
+	}
+
+	lchan->amr.codecs = n;
+	lchan->amr.dl_ft = start_codec;
+	lchan->amr.dl_cmr = start_codec;
+	lchan->amr.ul_ft = start_codec;
+	lchan->amr.ul_cmr = start_codec;
+
+	return 0;
+}
+
 int l1sched_activate_lchan(struct l1sched_ts *ts, enum l1sched_lchan_type chan)
 {
 	const struct l1sched_lchan_desc *lchan_desc = &l1sched_lchan_desc[chan];
 	struct l1sched_lchan_state *lchan;
 
 	/* Try to find requested logical channel */
-	lchan = l1sched_find_lchan(ts, chan);
+	lchan = l1sched_find_lchan_by_type(ts, chan);
 	if (lchan == NULL)
 		return -EINVAL;
 
@@ -492,6 +498,8 @@ int l1sched_activate_lchan(struct l1sched_ts *ts, enum l1sched_lchan_type chan)
 
 static void l1sched_reset_lchan(struct l1sched_lchan_state *lchan)
 {
+	struct msgb *msg;
+
 	/* Prevent NULL-pointer deference */
 	OSMO_ASSERT(lchan != NULL);
 
@@ -516,8 +524,9 @@ static void l1sched_reset_lchan(struct l1sched_lchan_state *lchan)
 	lchan->rx_bursts = NULL;
 	lchan->tx_bursts = NULL;
 
-	/* Forget the current prim */
-	l1sched_prim_drop(lchan);
+	/* Flush the queue of pending Tx prims */
+	while ((msg = msgb_dequeue(&lchan->tx_prims)) != NULL)
+		msgb_free(msg);
 
 	/* Channel specific stuff */
 	if (L1SCHED_CHAN_IS_TCH(lchan->type)) {
@@ -545,7 +554,7 @@ int l1sched_deactivate_lchan(struct l1sched_ts *ts, enum l1sched_lchan_type chan
 	struct l1sched_lchan_state *lchan;
 
 	/* Try to find requested logical channel */
-	lchan = l1sched_find_lchan(ts, chan);
+	lchan = l1sched_find_lchan_by_type(ts, chan);
 	if (lchan == NULL)
 		return -EINVAL;
 
@@ -608,35 +617,21 @@ enum gsm_phys_chan_config l1sched_chan_nr2pchan_config(uint8_t chan_nr)
 	return GSM_PCHAN_NONE;
 }
 
-enum l1sched_lchan_type l1sched_chan_nr2lchan_type(uint8_t chan_nr,
-	uint8_t link_id)
-{
-	int i;
-
-	/* Iterate over all known lchan types */
-	for (i = 0; i < _L1SCHED_CHAN_MAX; i++)
-		if (l1sched_lchan_desc[i].chan_nr == (chan_nr & 0xf8))
-			if (l1sched_lchan_desc[i].link_id == link_id)
-				return i;
-
-	return L1SCHED_IDLE;
-}
-
 static void l1sched_a5_burst_dec(struct l1sched_lchan_state *lchan,
-	uint32_t fn, sbit_t *burst)
+				 struct l1sched_burst_ind *bi)
 {
 	ubit_t ks[114];
 	int i;
 
 	/* Generate keystream for a DL burst */
-	osmo_a5(lchan->a5.algo, lchan->a5.key, fn, ks, NULL);
+	osmo_a5(lchan->a5.algo, lchan->a5.key, bi->fn, ks, NULL);
 
 	/* Apply keystream over ciphertext */
 	for (i = 0; i < 57; i++) {
 		if (ks[i])
-			burst[i + 3] *= -1;
+			bi->burst[i + 3] *= -1;
 		if (ks[i + 57])
-			burst[i + 88] *= -1;
+			bi->burst[i + 88] *= -1;
 	}
 }
 
@@ -709,7 +704,7 @@ static int subst_frame_loss(struct l1sched_lchan_state *lchan,
 		.toa256 = 0,
 		.rssi = -120,
 		.burst = { 0 },
-		.burst_len = GSM_BURST_LEN,
+		.burst_len = GSM_NBITS_NB_GMSK_BURST,
 	};
 
 	/* Traverse from fp till the current frame */
@@ -767,7 +762,7 @@ int l1sched_handle_rx_burst(struct l1sched_state *sched,
 		return -ENODEV;
 
 	/* Find required channel state */
-	lchan = l1sched_find_lchan(ts, chan);
+	lchan = l1sched_find_lchan_by_type(ts, chan);
 	if (lchan == NULL)
 		return -ENODEV;
 
@@ -782,7 +777,7 @@ int l1sched_handle_rx_burst(struct l1sched_state *sched,
 
 	/* Perform A5/X decryption if required */
 	if (lchan->a5.algo)
-		l1sched_a5_burst_dec(lchan, bi->fn, &bi->burst[0]);
+		l1sched_a5_burst_dec(lchan, bi);
 
 	/* Put burst to handler */
 	handler(lchan, bi);
@@ -824,7 +819,7 @@ int l1sched_handle_rx_probe(struct l1sched_state *sched,
 		return -ENODEV;
 
 	/* Find the appropriate logical channel */
-	lchan = l1sched_find_lchan(ts, frame->dl_chan);
+	lchan = l1sched_find_lchan_by_type(ts, frame->dl_chan);
 	if (lchan == NULL)
 		return -ENODEV;
 
