@@ -2,7 +2,7 @@
  * OsmocomBB <-> SDR connection bridge
  * The trxcon state machine (see 3GPP TS 44.004, section 5.1)
  *
- * (C) 2022 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
+ * (C) 2022-2023 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
  * Author: Vadim Yanitskiy <vyanitskiy@sysmocom.de>
  *
  * All Rights Reserved
@@ -29,13 +29,14 @@
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/logging.h>
 #include <osmocom/gsm/gsm0502.h>
+#include <osmocom/gsm/protocol/gsm_08_58.h>
 
 #include <osmocom/bb/trxcon/trxcon.h>
 #include <osmocom/bb/trxcon/trxcon_fsm.h>
 #include <osmocom/bb/trxcon/phyif.h>
 #include <osmocom/bb/trxcon/l1ctl.h>
 #include <osmocom/bb/l1sched/l1sched.h>
-#include <osmocom/bb/l1sched/logging.h>
+#include <osmocom/bb/l1gprs.h>
 
 #define S(x)	(1 << (x))
 
@@ -282,21 +283,25 @@ static void handle_tx_access_burst_req(struct osmo_fsm_inst *fi,
 				       const struct trxcon_param_tx_access_burst_req *req)
 {
 	struct trxcon_inst *trxcon = fi->priv;
-	enum l1sched_ts_prim_type prim_type;
-	const struct l1sched_ts_prim *prim;
+	struct l1sched_prim *prim;
+	struct msgb *msg;
 
-	const struct l1sched_ts_prim_rach rach = {
+	msg = l1sched_prim_alloc(L1SCHED_PRIM_T_RACH, PRIM_OP_REQUEST);
+	OSMO_ASSERT(msg != NULL);
+
+	prim = l1sched_prim_from_msgb(msg);
+	prim->rach_req = (struct l1sched_prim_rach) {
+		.chdr = {
+			.chan_nr = req->chan_nr,
+			.link_id = req->link_id,
+		},
 		.synch_seq = req->synch_seq,
 		.offset = req->offset,
+		.is_11bit = req->is_11bit,
 		.ra = req->ra,
 	};
 
-	prim_type = req->is_11bit ? L1SCHED_PRIM_RACH11 : L1SCHED_PRIM_RACH8;
-	prim = l1sched_prim_push(trxcon->sched, prim_type,
-				 req->chan_nr, req->link_id,
-				 (const uint8_t *)&rach, sizeof(rach));
-	if (prim == NULL)
-		LOGPFSML(fi, LOGL_ERROR, "Failed to enqueue a prim\n");
+	l1sched_prim_from_user(trxcon->sched, msg);
 }
 
 static void handle_dch_est_req(struct osmo_fsm_inst *fi,
@@ -364,6 +369,9 @@ static void handle_dch_est_req(struct osmo_fsm_inst *fi,
 		return;
 	}
 
+	/* Store TSC for subsequent PDCH timeslot activation(s) */
+	trxcon->l1p.tsc = req->tsc;
+
 	if (config == GSM_PCHAN_PDCH)
 		osmo_fsm_inst_state_chg(fi, TRXCON_ST_PACKET_DATA, 0, 0);
 	else
@@ -428,7 +436,7 @@ static void trxcon_st_dedicated_action(struct osmo_fsm_inst *fi,
 		break;
 	case TRXCON_EV_DCH_REL_REQ:
 		l1sched_reset(trxcon->sched, false);
-		osmo_fsm_inst_state_chg(fi, TRXCON_ST_RESET, 0, 0);
+		/* TODO: switch to (not implemented) TRXCON_ST_DCH_TUNING? */
 		break;
 	case TRXCON_EV_SET_TCH_MODE_REQ:
 	{
@@ -449,32 +457,14 @@ static void trxcon_st_dedicated_action(struct osmo_fsm_inst *fi,
 				/* Omit inactive channels */
 				if (!lchan->active)
 					continue;
-				lchan->tch_mode = req->mode;
 				if (req->mode == GSM48_CMODE_SPEECH_AMR) {
-					uint8_t bmask = req->amr.codecs_bitmask;
-					int n = 0;
-					int acum = 0;
-					int pos;
-					while ((pos = ffs(bmask)) != 0) {
-						acum += pos;
-						LOGPFSML(fi, LOGL_DEBUG,
-							 LOGP_LCHAN_NAME_FMT " AMR codec[%u] = %u\n",
-							 LOGP_LCHAN_NAME_ARGS(lchan), n, acum - 1);
-						lchan->amr.codec[n++] = acum - 1;
-						bmask >>= pos;
-					}
-					if (n == 0) {
-						LOGPFSML(fi, LOGL_ERROR,
-							 LOGP_LCHAN_NAME_FMT " Empty AMR codec mode bitmask!\n",
-							 LOGP_LCHAN_NAME_ARGS(lchan));
+					int rc = l1sched_lchan_set_amr_cfg(lchan,
+									   req->amr.codecs_bitmask,
+									   req->amr.start_codec);
+					if (rc)
 						continue;
-					}
-					lchan->amr.codecs = n;
-					lchan->amr.dl_ft = req->amr.start_codec;
-					lchan->amr.dl_cmr = req->amr.start_codec;
-					lchan->amr.ul_ft = req->amr.start_codec;
-					lchan->amr.ul_cmr = req->amr.start_codec;
 				}
+				lchan->tch_mode = req->mode;
 				req->applied = true;
 			}
 		}
@@ -502,15 +492,21 @@ static void trxcon_st_dedicated_action(struct osmo_fsm_inst *fi,
 	case TRXCON_EV_TX_DATA_REQ:
 	{
 		const struct trxcon_param_tx_data_req *req = data;
-		struct l1sched_ts_prim *prim;
+		struct l1sched_prim *prim;
+		struct msgb *msg;
 
-		prim = l1sched_prim_push(trxcon->sched, L1SCHED_PRIM_DATA,
-					 req->chan_nr, req->link_id,
-					 req->data, req->data_len);
-		if (prim == NULL) {
-			LOGPFSML(fi, LOGL_ERROR, "Failed to enqueue a prim\n");
-			return;
-		}
+		msg = l1sched_prim_alloc(L1SCHED_PRIM_T_DATA, PRIM_OP_REQUEST);
+		OSMO_ASSERT(msg != NULL);
+
+		prim = l1sched_prim_from_msgb(msg);
+		prim->data_req = (struct l1sched_prim_chdr) {
+			.chan_nr = req->chan_nr,
+			.link_id = req->link_id,
+			.traffic = req->traffic,
+		};
+
+		memcpy(msgb_put(msg, req->data_len), req->data, req->data_len);
+		l1sched_prim_from_user(trxcon->sched, msg);
 		break;
 	}
 	case TRXCON_EV_TX_DATA_CNF:
@@ -524,6 +520,55 @@ static void trxcon_st_dedicated_action(struct osmo_fsm_inst *fi,
 	}
 }
 
+static void handle_tbf_cfg_req(struct trxcon_inst *trxcon, uint8_t tn, bool active)
+{
+	struct l1sched_state *sched = trxcon->sched;
+
+	if (active) {
+		struct l1sched_lchan_state *lchan;
+		struct l1sched_ts *ts;
+
+		if (sched->ts[tn] != NULL) /* already enabled */
+			return;
+		if (l1sched_configure_ts(sched, tn, GSM_PCHAN_PDCH) != 0)
+			return;
+		OSMO_ASSERT(sched->ts[tn] != NULL);
+		ts = sched->ts[tn];
+
+		l1sched_activate_lchan(ts, L1SCHED_PDTCH);
+		l1sched_activate_lchan(ts, L1SCHED_PTCCH);
+		llist_for_each_entry(lchan, &ts->lchans, list)
+			lchan->tsc = trxcon->l1p.tsc;
+	} else {
+		l1sched_del_ts(sched, tn);
+	}
+}
+
+static void trxcon_l1gprs_state_changed_cb(struct l1gprs_pdch *pdch, bool active)
+{
+	handle_tbf_cfg_req(pdch->gprs->priv, pdch->tn, active);
+}
+
+static void trxcon_st_packet_data_onenter(struct osmo_fsm_inst *fi,
+					  uint32_t prev_state)
+{
+	struct trxcon_inst *trxcon = fi->priv;
+
+	OSMO_ASSERT(trxcon->gprs == NULL);
+	trxcon->gprs = l1gprs_state_alloc(trxcon, trxcon->log_prefix, trxcon);
+	l1gprs_state_set_pdch_changed_cb(trxcon->gprs, trxcon_l1gprs_state_changed_cb);
+	OSMO_ASSERT(trxcon->gprs != NULL);
+}
+
+static void trxcon_st_packet_data_onleave(struct osmo_fsm_inst *fi,
+					  uint32_t next_state)
+{
+	struct trxcon_inst *trxcon = fi->priv;
+
+	l1gprs_state_free(trxcon->gprs);
+	trxcon->gprs = NULL;
+}
+
 static void trxcon_st_packet_data_action(struct osmo_fsm_inst *fi,
 					 uint32_t event, void *data)
 {
@@ -536,14 +581,71 @@ static void trxcon_st_packet_data_action(struct osmo_fsm_inst *fi,
 	case TRXCON_EV_TX_ACCESS_BURST_CNF:
 		l1ctl_tx_rach_conf(trxcon, (const struct trxcon_param_tx_access_burst_cnf *)data);
 		break;
+	case TRXCON_EV_GPRS_UL_TBF_CFG_REQ:
+		l1gprs_handle_ul_tbf_cfg_req(trxcon->gprs, (struct msgb *)data);
+		break;
+	case TRXCON_EV_GPRS_DL_TBF_CFG_REQ:
+		l1gprs_handle_dl_tbf_cfg_req(trxcon->gprs, (struct msgb *)data);
+		break;
+	case TRXCON_EV_GPRS_UL_BLOCK_REQ:
+	{
+		struct l1gprs_prim_ul_block_req block_req;
+		struct l1sched_prim *prim;
+		struct msgb *msg = data;
+
+		if (l1gprs_handle_ul_block_req(trxcon->gprs, &block_req, msg) != 0)
+			return;
+
+		msg = l1sched_prim_alloc(L1SCHED_PRIM_T_DATA, PRIM_OP_REQUEST);
+		OSMO_ASSERT(msg != NULL);
+
+		prim = l1sched_prim_from_msgb(msg);
+		prim->data_req = (struct l1sched_prim_chdr) {
+			.frame_nr = block_req.hdr.fn,
+			.chan_nr = RSL_CHAN_OSMO_PDCH | block_req.hdr.tn,
+			.link_id = 0x00,
+		};
+
+		memcpy(msgb_put(msg, block_req.data_len), block_req.data, block_req.data_len);
+		l1sched_prim_from_user(trxcon->sched, msg);
+		break;
+	}
 	case TRXCON_EV_RX_DATA_IND:
 	{
 		const struct trxcon_param_rx_data_ind *ind = data;
+		struct l1gprs_prim_dl_block_ind block_ind;
+		struct msgb *msg;
+		uint8_t usf = 0xff;
 
-		if (ind->link_id == 0x00)
-			LOGPFSML(fi, LOGL_NOTICE, "Rx PDTCH/D message\n");
+		block_ind = (struct l1gprs_prim_dl_block_ind) {
+			.hdr = {
+				.fn = ind->frame_nr,
+				.tn = ind->chan_nr & 0x07,
+			},
+			.meas = {
+				/* .ber10k is set below */
+				.ci_cb = 180, /* 18 dB */
+				.rx_lev = dbm2rxlev(ind->rssi),
+			},
+			.data_len = ind->data_len,
+			.data = ind->data,
+		};
+
+		if (ind->n_bits_total == 0)
+			block_ind.meas.ber10k = 10000;
 		else
-			LOGPFSML(fi, LOGL_NOTICE, "Rx PTCCH/D message\n");
+			block_ind.meas.ber10k = 10000 * ind->n_errors / ind->n_bits_total;
+
+		msg = l1gprs_handle_dl_block_ind(trxcon->gprs, &block_ind, &usf);
+		if (msg != NULL)
+			trxcon_l1ctl_send(trxcon, msg);
+		/* Every fn % 13 == 12 we have either a PTCCH or an IDLE slot, thus
+		 * every fn % 13 ==  8 we add 5 frames, or 4 frames othrwise.  The
+		 * resulting value is first fn of the next block. */
+		const uint32_t rts_fn = GSM_TDMA_FN_SUM(ind->frame_nr, (ind->frame_nr % 13 == 8) ? 5 : 4);
+		msg = l1gprs_handle_rts_ind(trxcon->gprs, rts_fn, ind->chan_nr & 0x07, usf);
+		if (msg != NULL)
+			trxcon_l1ctl_send(trxcon, msg);
 		break;
 	}
 	case TRXCON_EV_DCH_EST_REQ:
@@ -551,7 +653,9 @@ static void trxcon_st_packet_data_action(struct osmo_fsm_inst *fi,
 		break;
 	case TRXCON_EV_DCH_REL_REQ:
 		l1sched_reset(trxcon->sched, false);
-		osmo_fsm_inst_state_chg(fi, TRXCON_ST_RESET, 0, 0);
+		/* TODO: switch to (not implemented) TRXCON_ST_DCH_TUNING? */
+		break;
+	case TRXCON_EV_TX_DATA_CNF:
 		break;
 	default:
 		OSMO_ASSERT(0);
@@ -569,6 +673,9 @@ static void trxcon_fsm_pre_term_cb(struct osmo_fsm_inst *fi,
 	/* Shutdown the scheduler */
 	if (trxcon->sched != NULL)
 		l1sched_free(trxcon->sched);
+	/* Clean up GPRS L1 state */
+	l1gprs_state_free(trxcon->gprs);
+
 	/* Close active connections */
 	if (trxcon->l2if != NULL)
 		trxcon_l1ctl_close(trxcon);
@@ -644,7 +751,13 @@ static const struct osmo_fsm_state trxcon_fsm_states[] = {
 				| S(TRXCON_EV_DCH_EST_REQ)
 				| S(TRXCON_EV_TX_ACCESS_BURST_REQ)
 				| S(TRXCON_EV_TX_ACCESS_BURST_CNF)
-				| S(TRXCON_EV_RX_DATA_IND),
+				| S(TRXCON_EV_GPRS_UL_TBF_CFG_REQ)
+				| S(TRXCON_EV_GPRS_DL_TBF_CFG_REQ)
+				| S(TRXCON_EV_GPRS_UL_BLOCK_REQ)
+				| S(TRXCON_EV_RX_DATA_IND)
+				| S(TRXCON_EV_TX_DATA_CNF),
+		.onenter = &trxcon_st_packet_data_onenter,
+		.onleave = &trxcon_st_packet_data_onleave,
 		.action = &trxcon_st_packet_data_action,
 	},
 };
@@ -670,6 +783,9 @@ static const struct value_string trxcon_fsm_event_names[] = {
 	OSMO_VALUE_STRING(TRXCON_EV_TX_DATA_CNF),
 	OSMO_VALUE_STRING(TRXCON_EV_RX_DATA_IND),
 	OSMO_VALUE_STRING(TRXCON_EV_CRYPTO_REQ),
+	OSMO_VALUE_STRING(TRXCON_EV_GPRS_UL_TBF_CFG_REQ),
+	OSMO_VALUE_STRING(TRXCON_EV_GPRS_DL_TBF_CFG_REQ),
+	OSMO_VALUE_STRING(TRXCON_EV_GPRS_UL_BLOCK_REQ),
 	{ 0, NULL }
 };
 
